import invariant from 'invariant';
import $$observable from 'symbol-observable';
import { from } from 'most';

function warning(message) {
  if (typeof console !== 'undefined' && typeof console.error === 'function') {
    console.error(message);
  }
  try {
    throw new Error(message);
  } catch (e) {
  }
}

function _defineProperty(obj, key, value) {
  if (key in obj) {
    Object.defineProperty(obj, key, {
      value: value,
      enumerable: true,
      configurable: true,
      writable: true
    });
  } else {
    obj[key] = value;
  }

  return obj;
}

function _objectSpread(target) {
  for (var i = 1; i < arguments.length; i++) {
    var source = arguments[i] != null ? arguments[i] : {};
    var ownKeys = Object.keys(source);

    if (typeof Object.getOwnPropertySymbols === 'function') {
      ownKeys = ownKeys.concat(Object.getOwnPropertySymbols(source).filter(function (sym) {
        return Object.getOwnPropertyDescriptor(source, sym).enumerable;
      }));
    }

    ownKeys.forEach(function (key) {
      _defineProperty(target, key, source[key]);
    });
  }

  return target;
}

function addToArray(a, b) {
  const i = a.indexOf(b);
  if (i === -1) {
    a.push(b);
  }
}
function removeFromArray(a, b) {
  const i = a.indexOf(b);
  if (i !== -1) {
    a.splice(i, 1);
  }
}
let _nextId = 0;
function nextId() {
  return _nextId++;
}
const unique = {
  equals(arg) {
    return false;
  },
  inspect() {
    return 'Unique{_}';
  }
};
function defaultEquals(a, b) {
  return Object.is(a, b) || a && typeof a.equals === 'function' && a.equals(b);
}
function setEquals(derivable, eq) {
  derivable._equals = eq;
  return derivable;
}
function equals(ctx, a, b) {
  return (ctx._equals || defaultEquals)(a, b);
}

const parentsStack = [];
let child = null;
function startCapturingParents(_child, parents) {
  parentsStack.push({
    parents,
    offset: 0,
    child: _child
  });
  child = _child;
}
function retrieveParentsFrame() {
  return parentsStack[parentsStack.length - 1];
}
function stopCapturingParents() {
  parentsStack.pop();
  child = parentsStack.length === 0 ? null : parentsStack[parentsStack.length - 1].child;
}
function maybeCaptureParent(p) {
  if (child === null) return;
  const frame = parentsStack[parentsStack.length - 1];
  if (frame.parents[frame.offset] === p) {
    frame.offset++;
    return;
  }
  const index = frame.parents.indexOf(p);
  if (index !== -1) {
    if (index <= frame.offset) return;
    const tmp = frame.parents[index];
    frame.parents[index] = frame.parents[frame.offset];
    frame.parents[frame.offset] = tmp;
    frame.offset++;
    return;
  }
  if (child !== undefined) {
    addToArray(p._activeChildren, child);
  }
  if (frame.offset === frame.parents.length) {
    frame.parents.push(p);
  } else {
    frame.parents.push(frame.parents[frame.offset]);
    frame.parents[frame.offset] = p;
  }
  frame.offset++;
}

const ATOM = 'ATOM';
const DERIVATION = 'DERIVATION';
const LENS = 'LENS';
const REACTOR = 'REACTOR';
function isDerivable(x) {
  return x !== undefined && x !== null && (x._type === DERIVATION || x._type === ATOM || x._type === LENS);
}

const UNKNOWN = 0;
const CHANGED = 1;
const UNCHANGED = 2;
const DISCONNECTED = 3;

function detach(parent, child) {
  removeFromArray(parent._activeChildren, child);
  if (parent._activeChildren.length === 0 && parent._parents != null) {
    const len = parent._parents.length;
    for (let i = 0; i < len; i++) {
      detach(parent._parents[i], parent);
    }
    parent._parents = null;
    parent._state = DISCONNECTED;
  }
}

class Reactor {
  constructor(parent, react) {
    this._governor = null;
    this._parent = void 0;
    this.react = void 0;
    this._active = false;
    this._reacting = false;
    this._parent = parent;
    this.react = react;
  }
  start() {
    this._active = true;
    addToArray(this._parent._activeChildren, this);
    this._parent.get();
  }
  _force(nextValue) {
    this._reacting = true;
    try {
      this.react(nextValue);
    } finally {
      this._reacting = false;
    }
  }
  force() {
    this._force(this._parent.get());
  }
  stop() {
    detach(this._parent, this);
    this._active = false;
  }
  _maybeReact() {
    if (!this._reacting && this._active) {
      if (this._governor !== null) {
        this._governor._maybeReact();
      }
      if (this._active) {
        const nextValue = this._parent.get();
        if (this._parent._state === CHANGED) {
          this._force(nextValue);
        }
      }
    }
  }
}
Object.defineProperty(Reactor.prototype, '_type', {
  value: REACTOR,
  configurable: true
});

function reactorFabric(derive, derivable, f, options) {
  !(typeof f === 'function') ? process.env.NODE_ENV !== "production" ? invariant(false, 'the first argument to .react must be a function') : invariant(false) : void 0;
  const opts = _objectSpread({
    once: false,
    skipFirst: false
  }, options);
  let skipFirst = opts.skipFirst;
  const reactor = new Reactor(derivable, val => {
    if (skipFirst) {
      skipFirst = false;
    } else {
      f(val);
      if (opts.once) {
        reactor.stop();
        controller.stop();
      }
    }
  });
  const assertCondition = (condition, name) => {
    if (isDerivable(condition)) {
      return condition;
    }
    if (typeof condition === 'function') {
      return condition;
    }
    if (typeof condition === 'undefined') {
      return condition;
    }
    throw Error(`react ${name} condition must be derivable or function, got: ${JSON.stringify(condition)}`);
  };
  function getCondition(condition, def) {
    if (!condition) return def;
    if (typeof condition === 'function') return condition(derivable);
    return condition.get();
  }
  const $from = assertCondition(opts.from, 'from');
  const $until = assertCondition(opts.until, 'until');
  const $when = assertCondition(opts.when, 'when');
  const $conds = derive(() => ({
    from: getCondition($from, true),
    until: getCondition($until, false),
    when: getCondition($when, true)
  }));
  let started = false;
  const controller = new Reactor($conds, conds => {
    if (conds.from) {
      started = true;
    }
    if (started) {
      if (conds.until) {
        reactor.stop();
        controller.stop();
      } else if (conds.when) {
        if (!reactor._active) {
          reactor.start();
          reactor.force();
        }
      } else if (reactor._active) {
        reactor.stop();
      }
    }
  });
  controller.start();
  controller.force();
  reactor._governor = controller;
}

function some(x) {
  return typeof x !== 'undefined' && x !== null;
}
function unpack(thing) {
  if (isDerivable(thing)) {
    return thing.get();
  } else {
    return thing;
  }
}
class Derivation {
  constructor(deriver, meta = null) {
    this._parents = null;
    this._value = unique;
    this._equals = null;
    this._state = DISCONNECTED;
    this._deriver = void 0;
    this._meta = void 0;
    this._activeChildren = [];
    this._deriver = deriver;
    this._meta = meta;
  }
  _clone() {
    return setEquals(derive(this._deriver), this._equals);
  }
  _forceEval() {
    let newVal = null;
    let newNumParents;
    try {
      if (this._parents === null) {
        this._parents = [];
      }
      startCapturingParents(this, this._parents);
      newVal = this._deriver();
      newNumParents = retrieveParentsFrame().offset;
    } finally {
      stopCapturingParents();
    }
    if (!equals(this, newVal, this._value)) {
      this._state = CHANGED;
    } else {
      this._state = UNCHANGED;
    }
    for (let i = newNumParents, len = this._parents.length; i < len; i++) {
      const oldParent = this._parents[i];
      detach(oldParent, this);
      this._parents[i] = null;
    }
    this._parents.length = newNumParents;
    this._value = newVal;
  }
  _update() {
    if (this._parents === null) {
      this._forceEval();
    } else if (this._state === UNKNOWN) {
      const len = this._parents.length;
      for (let i = 0; i < len; i++) {
        const parent = this._parents[i];
        if (parent._state === UNKNOWN) {
          parent._update();
        }
        if (parent._state === CHANGED) {
          this._forceEval();
          break;
        }
      }
      if (this._state === UNKNOWN) {
        this._state = UNCHANGED;
      }
    }
  }
  get() {
    maybeCaptureParent(this);
    if (this._activeChildren.length > 0) {
      this._update();
    } else {
      startCapturingParents(void 0, []);
      try {
        this._value = this._deriver();
      } finally {
        stopCapturingParents();
      }
    }
    return this._value;
  }
  derive(f) {
    !(typeof f === 'function') ? process.env.NODE_ENV !== "production" ? invariant(false, 'derive requires function') : invariant(false) : void 0;
    return derive(() => f(this.get()));
  }
  maybeDerive(f) {
    !(typeof f === 'function') ? process.env.NODE_ENV !== "production" ? invariant(false, 'maybeDerive requires function') : invariant(false) : void 0;
    return derive(() => {
      const arg = this.get();
      return some(arg) ? f(arg) : null;
    });
  }
  orDefault(def) {
    !some(def) ? process.env.NODE_ENV !== "production" ? invariant(false, 'orDefault requires non-null value') : invariant(false) : void 0;
    return this.derive(value => some(value) ? value : def);
  }
  react(f, opts) {
    reactorFabric(derive, this, f, opts);
  }
  maybeReact(f, opts) {
    let maybeWhen = this.derive(Boolean);
    if (opts && 'when' in opts && opts.when !== true) {
      let when = opts.when;
      if (typeof when === 'function' || when === false) {
        when = derive(when);
      }
      !isDerivable(when) ? process.env.NODE_ENV !== "production" ? invariant(false, 'when condition must be bool, function, or derivable') : invariant(false) : void 0;
      maybeWhen = maybeWhen.derive(d => d && when.get());
    }
    reactorFabric(derive, this, f, _objectSpread({}, opts, {
      when: maybeWhen
    }));
  }
  is(other) {
    return derive(() => equals(this, this.get(), unpack(other)));
  }
  withEquality(equals$$1) {
    let selectedEquals = equals$$1;
    if (equals$$1) {
      !(typeof equals$$1 === 'function') ? process.env.NODE_ENV !== "production" ? invariant(false, 'equals must be function') : invariant(false) : void 0;
    } else {
      selectedEquals = null;
    }
    return setEquals(this._clone(), selectedEquals);
  }
}
Object.defineProperty(Derivation.prototype, '_type', {
  value: DERIVATION,
  configurable: true
});
function derive(f, meta) {
  !(typeof f === 'function') ? process.env.NODE_ENV !== "production" ? invariant(false, 'derive requires function') : invariant(false) : void 0;
  return new Derivation(f, meta);
}

function mark(node, reactors) {
  for (let i = 0, len = node._activeChildren.length; i < len; i++) {
    const child = node._activeChildren[i];
    switch (child._type) {
      case DERIVATION:
      case LENS:
        if (child._state !== UNKNOWN) {
          child._state = UNKNOWN;
          mark(child, reactors);
        }
        break;
      case REACTOR:
        reactors.push(child);
        break;
    }
  }
}
function processReactors(reactors) {
  for (let i = 0, len = reactors.length; i < len; i++) {
    const r = reactors[i];
    if (r._reacting) {
      throw Error('Synchronous cyclical reactions disallowed. ' + 'Use setImmediate.');
    }
    r._maybeReact();
  }
}
const TransactionAbortion = {};
function initiateAbortion() {
  throw TransactionAbortion;
}
class TransactionContext {
  constructor(parent) {
    this.id2originalValue = {};
    this.parent = void 0;
    this.modifiedAtoms = [];
    this.parent = parent;
  }
}
function maybeTrack(atom) {
  if (currentCtx === null) return;
  if (atom._id in currentCtx.id2originalValue) return;
  currentCtx.modifiedAtoms.push(atom);
  currentCtx.id2originalValue[atom._id] = atom._value;
}
let currentCtx = null;
function inTransaction() {
  return currentCtx !== null;
}
function transact(f) {
  beginTransaction();
  try {
    f(initiateAbortion);
  } catch (e) {
    abortTransaction();
    if (e !== TransactionAbortion) {
      throw e;
    }
    return;
  }
  commitTransaction();
}
function atomically(f) {
  if (!inTransaction()) {
    transact(f);
  } else {
    f();
  }
}
function beginTransaction() {
  currentCtx = new TransactionContext(currentCtx);
}
function commitTransaction() {
  if (currentCtx === null) return;
  const ctx = currentCtx;
  currentCtx = ctx.parent;
  if (currentCtx !== null) return;
  const reactors = [];
  ctx.modifiedAtoms.forEach(a => {
    if (equals(a, a._value, ctx.id2originalValue[a._id])) {
      a._state = UNCHANGED;
    } else {
      a._state = CHANGED;
      mark(a, reactors);
    }
  });
  processReactors(reactors);
  ctx.modifiedAtoms.forEach(a => {
    a._state = UNCHANGED;
  });
}
function abortTransaction() {
  if (currentCtx === null) return;
  const ctx = currentCtx;
  currentCtx = ctx.parent;
  ctx.modifiedAtoms.forEach(atom => {
    atom._value = ctx.id2originalValue[atom._id];
    atom._state = UNCHANGED;
    mark(atom, []);
  });
}

function makeReactor(derivable, f, options) {
  return reactorFabric(derive, derivable, f, options);
}

function update(instance, f, args) {
  switch (args.length) {
    case 0:
      return instance.set(f(instance.get()));
    case 1:
      return instance.set(f(instance.get(), args[0]));
    case 2:
      return instance.set(f(instance.get(), args[0], args[1]));
    case 3:
      return instance.set(f(instance.get(), args[0], args[1], args[2]));
    case 4:
      return instance.set(f(instance.get(), args[0], args[1], args[2], args[3]));
    default:
      throw Error('update function accepts only 4 arguments');
  }
}

function unpack$1(thing) {
  if (isDerivable(thing)) {
    return thing.get();
  } else {
    return thing;
  }
}

function some$1(x) {
  return typeof x !== 'undefined' && x !== null;
}
class Atom {
  constructor(value, meta = null) {
    this._id = nextId();
    this._state = UNCHANGED;
    this._equals = null;
    this._value = void 0;
    this._meta = void 0;
    this._activeChildren = [];
    this._value = value;
    this._meta = meta;
  }
  derive(f) {
    !(typeof f === 'function') ? process.env.NODE_ENV !== "production" ? invariant(false, 'derive requires function') : invariant(false) : void 0;
    return derive(() => f(this.get()));
  }
  maybeDerive(f) {
    !(typeof f === 'function') ? process.env.NODE_ENV !== "production" ? invariant(false, 'maybeDerive requires function') : invariant(false) : void 0;
    return derive(() => {
      const arg = this.get();
      return some$1(arg) ? f(arg) : null;
    });
  }
  orDefault(def) {
    !some$1(def) ? process.env.NODE_ENV !== "production" ? invariant(false, 'orDefault requires non-null value') : invariant(false) : void 0;
    return this.derive(value => some$1(value) ? value : def);
  }
  react(f, opts) {
    makeReactor(this, f, opts);
  }
  maybeReact(f, opts) {
    let maybeWhen = this.derive(Boolean);
    if (opts && 'when' in opts && opts.when !== true) {
      let when = opts.when;
      if (typeof when === 'function' || when === false) {
        when = derive(when);
      }
      !isDerivable(when) ? process.env.NODE_ENV !== "production" ? invariant(false, 'when condition must be bool, function, or derivable') : invariant(false) : void 0;
      maybeWhen = maybeWhen.derive(d => d && when && when.get());
    }
    makeReactor(this, f, _objectSpread({}, opts, {
      when: maybeWhen
    }));
  }
  is(other) {
    return derive(() => equals(this, this.get(), unpack$1(other)));
  }
  withEquality(equals$$1) {
    let eq = equals$$1;
    if (eq) {
      !(typeof equals$$1 === 'function') ? process.env.NODE_ENV !== "production" ? invariant(false, 'equals must be function') : invariant(false) : void 0;
    } else {
      eq = null;
    }
    return setEquals(this._clone(), eq);
  }
  _clone() {
    return setEquals(atom(this._value), this._equals);
  }
  set(value) {
    maybeTrack(this);
    const oldValue = this._value;
    this._value = value;
    if (!inTransaction()) {
      if (!equals(this, value, oldValue)) {
        try {
          this._state = CHANGED;
          const reactors = [];
          mark$1(this, reactors);
          processReactors(reactors);
        } finally {
          this._state = UNCHANGED;
        }
      }
    }
  }
  get() {
    if (typeof global.__DERIVABLE_DEVTOOLS_HOOK__ === 'function') {
      global.__DERIVABLE_DEVTOOLS_HOOK__('captureAtom', this);
    }
    maybeCaptureParent(this);
    return this._value;
  }
  update(f, ...args) {
    return update(this, f, args);
  }
}
Object.defineProperty(Atom.prototype, '_type', {
  value: ATOM,
  configurable: true
});
function atom(value, meta) {
  return new Atom(value, meta);
}
function mark$1(node, reactors) {
  for (let i = 0, len = node._activeChildren.length; i < len; i++) {
    const child = node._activeChildren[i];
    switch (child._type) {
      case DERIVATION:
      case LENS:
        if (child._state !== UNKNOWN) {
          child._state = UNKNOWN;
          mark$1(child, reactors);
        }
        break;
      case REACTOR:
        reactors.push(child);
        break;
    }
  }
}

class Lens extends Derivation {
  constructor(descriptor, meta) {
    super(descriptor.get, meta);
    this._descriptor = void 0;
    this._descriptor = descriptor;
  }
  _clone() {
    return setEquals(new Lens(this._descriptor), this._equals);
  }
  set(value) {
    atomically(() => {
      this._descriptor.set(value);
    });
  }
  update(f, ...args) {
    return update(this, f, args);
  }
}
Object.defineProperty(Lens.prototype, '_type', {
  value: LENS,
  configurable: true
});

function createEvent(name) {
  return eventConstructor({
    name,
    domainName: ''
  });
}
function createEffect(name) {
  return effectConstructor({
    name,
    domainName: ''
  });
}
function eventConstructor({
  name,
  domainName
}) {
  const fullName = makeName(name, domainName);
  const eventState = atom({
    payload: null
  });
  const instance = payload => instance.create(payload, fullName);
  setProperty('create', create, instance);
  setProperty('eventState', eventState, instance);
  setProperty('toString', getType, instance);
  setProperty('getType', getType, instance);
  setProperty('kind', () => 'event', instance);
  setProperty($$observable, () => instance, instance);
  instance.watch = watch;
  instance.map = map;
  instance.subscribe = subscribe;
  instance.to = to;
  instance.epic = epic;
  function epic(fn) {
    const instance$ = from(instance).multicast();
    const epic$ = fn(instance$).multicast();
    const mapped = eventConstructor({
      name: `${name}$ ~> *`,
      domainName
    });
    epic$.observe(mapped);
    return mapped;
  }
  function to(target, handler) {
    !(target.kind() === 'store') ? process.env.NODE_ENV !== "production" ? invariant(false, 'right now event.to support only stores') : invariant(false) : void 0;
    !(typeof target.setState === 'function') ? process.env.NODE_ENV !== "production" ? invariant(false, 'right now event.to support only stores') : invariant(false) : void 0;
    watch(payload => target.setState(payload, handler));
  }
  function create(payload, fullName) {
    instance.eventState.set({
      payload
    });
    return payload;
  }
  function watch(watcher) {
    instance.eventState.react(() => {
      watcher(instance.eventState.get().payload);
    }, {
      skipFirst: true
    });
  }
  function subscribe(observer) {
    watch(payload => observer.next(payload));
    function unsubscribe() {
      process.env.NODE_ENV !== "production" ? warning('TODO event.unsubscribe is not implemented') : void 0;
    }
    unsubscribe.unsubscribe = unsubscribe;
    return unsubscribe;
  }
  function map(fn) {
    const mapped = eventConstructor({
      name: `${name} → *`,
      domainName
    });
    watch(payload => {
      mapped(fn(payload));
    });
    return mapped;
  }
  function getType() {
    return fullName;
  }
  return instance;
}
function effectConstructor({
  name,
  domainName
}) {
  const instance = eventConstructor({
    name,
    domainName
  });
  const eventCreate = instance.create;
  const done = eventConstructor({
    name: name + ' done',
    domainName
  });
  const fail = eventConstructor({
    name: name + ' fail',
    domainName
  });
  instance.done = done;
  instance.fail = fail;
  instance.use = use;
  setProperty('create', create, instance);
  setProperty('kind', () => 'effect', instance);
  let thunk = async value => {
    process.env.NODE_ENV !== "production" ? warning('no thunk used') : void 0;
  };
  function use(fn) {
    thunk = fn;
  }
  function create(payload) {
    const throwSymbol = {};
    let syncError = throwSymbol;
    let req;
    try {
      req = thunk(payload);
    } catch (err) {
      syncError = err;
    }
    const isSyncError = syncError !== throwSymbol;
    const isPromise = !isSyncError && hasPromise(req);
    if (isPromise) req.then(done, fail);else if (isSyncError) fail(syncError);else done(req);
    eventCreate(payload, instance.getType());
    return {
      done() {
        return new Promise(rs => {
          if (isPromise) req.then(rs);else if (!isSyncError) rs(req);
        });
      },
      fail() {
        return new Promise(rs => {
          if (isPromise) req.catch(rs);else if (isSyncError) rs(syncError);
        });
      },
      promise() {
        if (isPromise) return req;
        if (isSyncError) return Promise.reject(syncError);
        return Promise.resolve(req);
      }
    };
  }
  return instance;
}
function hasPromise(req) {
  return typeof req === 'object' && req !== null && typeof req.then === 'function';
}
function makeName(name, domainName) {
  return [name, domainName].filter(str => str.length > 0).join('/');
}
function setProperty(prop, value, obj) {
  Object.defineProperty(obj, prop, {
    value,
    writable: true,
    configurable: true
  });
}

const INIT = '@@redux/INIT' + Math.random().toString(36).substring(7).split('').join('.');
const REPLACE = '@@redux/REPLACE' + Math.random().toString(36).substring(7).split('').join('.');

function compose(...funcs) {
  if (funcs.length === 0) {
    return arg => arg;
  }
  if (funcs.length === 1) {
    return funcs[0];
  }
  return funcs.reduce((a, b) => (...args) => a(b(...args)));
}

function applyMiddleware(...middlewares) {
  return createStore => (...args) => {
    const store = createStore(...args);
    let dispatch = () => {
      throw new Error(`Dispatching while constructing your middleware is not allowed. ` + `Other middleware would not be applied to this dispatch.`);
    };
    let chain = [];
    const middlewareAPI = {
      getState: store.getState,
      dispatch: (...args) => dispatch(...args)
    };
    chain = middlewares.map(middleware => middleware(middlewareAPI));
    dispatch = compose(...chain)(store.dispatch);
    return _objectSpread({}, store, {
      dispatch
    });
  };
}

function* untilEnd(set) {
  do {
    for (const e of set) {
      set.delete(e);
      yield e;
    }
  } while (set.size > 0);
}
let id = 0;
function storeConstructor(props) {
  const currentId = (++id).toString(36);
  let currentListeners = [];
  const pending = new Set();
  let isDispatching = false;
  let {
    currentReducer,
    currentState
  } = props;
  let nextListeners = currentListeners;
  const nests = getNested(currentState, fn => {
    if (!isDispatching) return setState(fn(currentState));
    pending.add(fn);
  });
  let defaultState = currentState;
  let needToSaveFirst = defaultState === undefined;
  nests.add({
    get() {
      return getState();
    },
    set(state, action) {
      return currentReducer(state, action);
    }
  });
  function ensureCanMutateNextListeners() {
    if (nextListeners === currentListeners) {
      nextListeners = currentListeners.slice();
    }
  }
  function getState() {
    !!isDispatching ? process.env.NODE_ENV !== "production" ? invariant(false, 'You may not call store.getState() while the reducer is executing. ' + 'The reducer has already received the state as an argument. ' + 'Pass it down from the top reducer instead of reading it from the store.') : invariant(false) : void 0;
    return currentState;
  }
  function subscribe(listener) {
    !(typeof listener === 'function') ? process.env.NODE_ENV !== "production" ? invariant(false, 'Expected the listener to be a function.') : invariant(false) : void 0;
    !!isDispatching ? process.env.NODE_ENV !== "production" ? invariant(false, 'You may not call store.subscribe() while the reducer is executing. ' + 'If you would like to be notified after the store has been updated, subscribe from a ' + 'component and invoke store.getState() in the callback to access the latest state') : invariant(false) : void 0;
    let isSubscribed = true;
    ensureCanMutateNextListeners();
    nextListeners.push(listener);
    return function unsubscribe() {
      if (!isSubscribed) {
        return;
      }
      !!isDispatching ? process.env.NODE_ENV !== "production" ? invariant(false, 'You may not unsubscribe from a store listener while the reducer is executing') : invariant(false) : void 0;
      isSubscribed = false;
      ensureCanMutateNextListeners();
      const index = nextListeners.indexOf(listener);
      nextListeners.splice(index, 1);
    };
  }
  function dispatch(action) {
    !(typeof action.type !== 'undefined') ? process.env.NODE_ENV !== "production" ? invariant(false, 'Actions may not have an undefined "type" property.') : invariant(false) : void 0;
    !!isDispatching ? process.env.NODE_ENV !== "production" ? invariant(false, 'Reducers may not dispatch actions.') : invariant(false) : void 0;
    isDispatching = true;
    try {
      currentState = setNested(currentState, action, nests);
      for (const fn of untilEnd(pending)) {
        currentState = fn(currentState);
      }
    } finally {
      if (needToSaveFirst === true && currentState !== undefined) {
        needToSaveFirst = false;
        defaultState = currentState;
      }
      isDispatching = false;
    }
    const listeners = currentListeners = nextListeners;
    for (let i = 0; i < listeners.length; i++) {
      const listener = listeners[i];
      listener(currentState, action);
    }
    return action;
  }
  function replaceReducer(nextReducer) {
    !(typeof nextReducer === 'function') ? process.env.NODE_ENV !== "production" ? invariant(false, 'Expected the nextReducer to be a function.') : invariant(false) : void 0;
    currentReducer = nextReducer;
    dispatch({
      type: REPLACE
    });
  }
  function observable() {
    const outerSubscribe = subscribe;
    return {
      subscribe(observer) {
        !(typeof observer === 'object' && observer !== null) ? process.env.NODE_ENV !== "production" ? invariant(false, 'Expected the observer to be an object.') : invariant(false) : void 0;
        function observeState() {
          if (observer.next) {
            observer.next(getState());
          }
        }
        observeState();
        const unsubscribe = outerSubscribe(observeState);
        return {
          unsubscribe
        };
      },
      [$$observable]() {
        return this;
      }
    };
  }
  function reset(event) {
    on(event, () => {
      setState(defaultState);
    });
    return store;
  }
  function on(event, handler) {
    event.to(store, handler);
    return store;
  }
  function withProps(fn) {
    return props => fn(getState(), props);
  }
  function map(fn) {
    const innerStore = createStore(fn(getState()));
    const unsub = watch(update => {
      const mapped = fn(update);
      innerStore.setState(mapped);
    });
    return innerStore;
  }
  function to(action, reduce) {
    const needReduce = action.kind() === 'store' && typeof reduce === 'function';
    return watch(data => {
      if (!needReduce) {
        action(data);
      } else {
        const lastState = action.getState();
        const reduced = reduce(lastState, data);
        if (lastState !== reduced) action.setState(reduced);
      }
    });
  }
  function watch(fn) {
    return subscribe(() => {
      fn(getState());
    });
  }
  function epic(fn) {
    return epicStore(store, fn);
  }
  function stateSetter(_, payload) {
    return payload;
  }
  function setState(value, reduce) {
    const currentReducer = typeof reduce === 'function' ? reduce : stateSetter;
    currentState = currentReducer(currentState, value);
    dispatch({
      type: `set state ${currentId}`,
      payload: value
    });
  }
  dispatch({
    type: INIT
  });
  const store = {
    id: currentId,
    withProps,
    setState,
    dispatch,
    map,
    on,
    to,
    watch,
    epic,
    subscribe,
    getState,
    replaceReducer,
    reset,
    [$$observable]: observable
  };
  Object.defineProperty(store, 'kind', {
    writable: true,
    configurable: true,
    value() {
      return 'store';
    }
  });
  return store;
}
function epicStore(store, fn) {
  const store$ = from(store).multicast();
  const mapped$ = fn(store$).multicast();
  const innerStore = createStore();
  const subs = mapped$.subscribe({
    next(value) {
      innerStore.setState(value);
    },
    error(err) {
      if (process.env.NODE_ENV !== "production") console.error(err);
    },
    complete() {
      subs();
    }
  });
  return innerStore;
}
function hasKind(value) {
  return typeof value === 'object' && value !== null && typeof value.kind === 'function';
}
function createStore(preloadedStateRaw, reducer = _ => _, enhancerRaw) {
  let enhancer = enhancerRaw;
  let preloadedState = preloadedStateRaw;
  if (typeof preloadedState === 'function' && typeof enhancer === 'undefined') {
    enhancer = preloadedState;
    preloadedState = undefined;
  }
  if (typeof enhancer !== 'undefined') {
    if (typeof enhancer !== 'function') {
      !Array.isArray(enhancer) ? process.env.NODE_ENV !== "production" ? invariant(false, 'Expected the enhancer to be a function') : invariant(false) : void 0;
      enhancer = applyMiddleware(...enhancer);
    }
    return enhancer(createStore)(reducer, preloadedState);
  }
  !(typeof reducer === 'function') ? process.env.NODE_ENV !== "production" ? invariant(false, 'Expected the reducer to be a function') : invariant(false) : void 0;
  return storeConstructor({
    currentReducer: reducer,
    currentState: preloadedState
  });
}
function setNested(initialState, action, nests) {
  let currentState = initialState;
  for (const nest of nests) {
    currentState = nest.set(currentState, action);
  }
  return currentState;
}
function nest(value, key) {
  return {
    get() {
      return value.getState();
    },
    set(state, action) {
      value.dispatch(action);
      state[key] = value.getState();
      return state;
    }
  };
}
function getNested(initialState, setState) {
  const nests = new Set();
  if (typeof initialState !== 'object' || initialState === null) return nests;
  for (const [key, value] of Object.entries(_objectSpread({}, initialState))) {
    if (hasKind(value) && value.kind() === 'store') {
      const n = nest(value, key);
      nests.add(n);
      value.watch(e => {
        setState(state => _objectSpread({}, state, {
          [key]: e
        }));
      });
      initialState[key] = value.getState();
    }
  }
  return nests;
}

function createDomain({
  domain
}) {
  return {
    event(name) {
      return createEvent(name);
    },
    effect(name) {
      return createEffect(name);
    },
    domain(name) {
      return createDomain({
        domain: [name].concat(domain)
      });
    },
    store(state) {
      return createStore(state);
    }
  };
}
function createRootDomain(name) {
  if (name === undefined) return createDomain({
    domain: []
  });
  return createDomain({
    domain: [name]
  });
}

const defaultDomain = createRootDomain();
function createEvent$1(name) {
  return defaultDomain.event(name);
}
function createEffect$1(name) {
  return defaultDomain.effect(name);
}

export { createRootDomain as createDomain, createStore, createEvent$1 as createEvent, createEffect$1 as createEffect };
//# sourceMappingURL=effector.es.js.map
